#!/usr/bin/env python3

# Green Pass Parser
# Copyright (C) 2021  Davide Berardi -- <berardi.dav@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Base url to retrieve data (DGC)
BASE_URL_DGC = "https://get.dgc.gov.it/v1/dgc/"
# Alternative base url to retrieve data (DGC)
BASE_URL_DGCG = "https://dgcg.covidbevis.se/tp/"
# Base url to retrieve data (NHS)
BASE_URL_NHS = "https://covid-status.service.nhsx.nhs.uk/"
